"""service_bus

This module implements various functionality related to Azure Service Bus.

Author:
    Sam Gibson <sgibson@glasswallsolutions.com>
"""
import logging
from typing import List
from os.path import join
import pickle

from azure.servicebus import ServiceBusClient
from azure.servicebus import Message
from azure.servicebus import ReceiveSettleMode


def get_all_dead_letter_ids(queue_str: str,
                            client: ServiceBusClient) -> List[str]:
    """Get a list of all message IDs in the dead letter queue of a given
    queue.

    Args:
        queue_str: The name of the queue.
        client: The service bus client.
    """
    logging.info(f"--> Getting dead letters from '{queue_str}'...")
    dead_letters = []
    queue = client.get_queue(queue_str)
    with queue.get_deadletter_receiver(idle_timeout=0.2) as dead_letter_rx:
        dead_letters += dead_letter_rx.peek(count=9999)

    dead_letter_ids = []
    for msg in dead_letters:
        dead_letter_ids.append(msg.properties.message_id.decode("utf-8"))
    logging.info(
        f"    Finished scanning. {len(dead_letters)} dead letter(s) were found."
    )
    return dead_letter_ids


def connect(connection_str: str) -> ServiceBusClient:
    """Initiate a connection to a service bus with a connection string."""
    return ServiceBusClient.from_connection_string(connection_str)


def copy_message(msg: Message) -> Message:
    """Create a copy of the service bus message given."""
    msg_body = b''
    for b in msg.body:
        msg_body += b
    new_msg = Message(msg_body)
    new_msg.properties = msg.properties
    return new_msg


def store_message_in_vault(message: Message, vault_dir: str):
    """Store a message's body in the vault in the event of failure.

    Args:
        message (Message): The message to store.
        vault_dir (str): The path to the vault folder to store it in.
                         This is 'vault_dir/cluster_name/queue_name/'.
    """
    # get the path to the file
    vault_file_path = join(vault_dir,
                           message.properties.message_id.decode("utf-8"))

    # write the message body
    with open(vault_file_path, 'wb') as vault_file:
        body = b''
        for b in message.body:
            body += b
        vault_file.write(body)


def resend_dead_letters_from_queue(vault_dir: str, queue: str,
                                   client: ServiceBusClient) -> List[str]:
    """Scan a queue for dead letters, and resend found.

    If there are any errors resending, store the message body in the vault.

    Args:
        vault_dir (str): The root dir of the vault.
        queue (str): The name of the queue to scan.
        client (ServiceBusClient): The service bus client to use.

    Returns:
        List[str]: The list of transaction IDs that were resent.
    """
    logging.info(f"--> Getting dead letters from '{queue}'...")
    resent_ids = []
    queue = client.get_queue(queue)
    with queue.get_deadletter_receiver(idle_timeout=0.2) as dead_letter_rx:
        with queue.get_sender() as queue_sender:
            number_replayed = 0

            # receive the messages from the dead letter queue
            for msg in dead_letter_rx:
                message_id = msg.properties.message_id.decode('utf-8')

                # messages without a body are invalid, we should skip those
                if msg.body is None:
                    logging.warn(
                        f"\tMessage '{message_id}' did not have a body.")
                    msg.complete()
                    continue

                try:
                    # now attempt to replay the error
                    logging.info(
                        f"\tReplaying a dead letter ({message_id})...")
                    store_message_in_vault(msg, vault_dir)
                    to_send = copy_message(msg)
                    queue_sender.send(to_send)
                    number_replayed += 1
                    resent_ids.append(message_id)
                except:
                    # if we get any errors while replaying, make sure the message
                    # is abandoned, and kept on the dead letter queue
                    logging.error("--> An error occurred while processing a dead letter."\
                        " Abandoning the message.")
                    msg.abandon()
                    return
                msg.complete()
            logging.info(
                f"    Finished scanning. {number_replayed} dead letter(s) were found and replayed."
            )
    return resent_ids



from io import BytesIO
import json
import logging
import os

from azure.common import AzureMissingResourceHttpError
from azure.core.paging import ItemPaged
from azure.storage.blob import BlobServiceClient, BlobProperties, ContainerClient

# sometimes it's in different places in the blob storage container, not
# sure what causes this or even if it's still an issue but in November-ish 2019
# we had reconstructions that weren't working due to Received/MimeMessage not
# being present - so we always need to check for both of these
MIME_BLOB_NAMES = [
    "Received/MimeMessage",
    "MessageInspectionQueue/Glasswall.FileTrust.Messaging.ReceivedMessage.json"
]
"""The name of the blob within the container that contains the MIME message."""

CONNECTION_STR = os.environ.get('AZURE_STORAGE_CONNECTION_STRING')


def connect(connection_str: str) -> BlobServiceClient:
    """Connect to the Azure Blob Storage account using a StorageAccount.

    Args:
        storage_account (StorageAccount): The storage account creds to use.

    Returns:
        BlobServiceClient: The blob service to use.
    """
    return BlobServiceClient.from_connection_string(connection_str)


def get_mime_message(transaction_id: str,
                     blob_service: BlobServiceClient) -> str:
    """Get the MIME message received from a transaction ID.

    Args:
        transaction_id (str): The transaction ID to use to get the message.
        blob_service (BlockBlobService): The service to use to get the message.

    Returns:
        str: The MIME message.
    """
    for blob_name in MIME_BLOB_NAMES:
        try:
            logging.debug(transaction_id, blob_name)
            blob_client = blob_service.get_blob_client(transaction_id.lower(),
                                                       blob_name)
            blob = blob_client.download_blob()
            blob_content = BytesIO()
            blob_content.write(blob.readall())
            blob_content = blob_content.getvalue().decode('utf-8')

            if blob_name.endswith(".json"):
                mime_json = json.loads(blob_content)
                return mime_json["receivedMimeMessage"]
            else:
                return blob_content
        except AzureMissingResourceHttpError:
            # this is pure filth, but some of the blobs are in messed up formats
            # and we need to try multiple times to get the MIME message
            continue

    logging.error(
        f"Transaction ID '{transaction_id}' not found in blob storage")
    return None


def get_container(container_name: str, connection_str: str) -> ContainerClient:
    return connect(connection_str).get_container_client(container_name)


def get_blob_properties(container_name: str, connection_str: str) -> ItemPaged[BlobProperties]:
    return get_container(container_name, connection_str).list_blobs()



"""send_mail

Implements the mailsend.py functionality in the Victoria plugin.

Author:
    Sam Gibson <sgibson@glasswallsolutions.com>
"""
from __future__ import annotations
from datetime import datetime
import logging
from os import path
from typing import List
import uuid

from marshmallow import Schema, fields, post_load, validate, ValidationError, EXCLUDE
from sremail import address, message, smtp
import yaml


class ManifestSchema(Schema):
    """Marshmallow schema for an email manifest."""
    sender = address.AddressField(required=True, allow_none=False)
    to = fields.List(address.AddressField,
                     required=True,
                     validate=validate.Length(min=1),
                     allow_none=False)
    smtp_server = fields.String(required=True, allow_none=False)
    tenant_id = fields.UUID(required=True)
    port = fields.Int(missing=25, allow_none=False)
    attach = fields.List(fields.Str, missing=[], allow_none=False)

    @post_load
    def make_manifest(self, data, **kwargs):
        return Manifest(**data)


MANIFEST_SCHEMA = ManifestSchema(unknown=EXCLUDE)
"""A module-wide instance of ManifestSchema to use when
loading a Manifest."""


class Manifest:
    """A manifest is a description of an email we want to send.

    Attributes:
        sender: The email address sending the email.
        to: The list of emails to send the email to.
        smtp_server: The SMTP server to send the email to.
        tenant_id: The FileTrust tenant ID to send the email with.
        port: The port on the SMTP server to connect via.
        attach: The list of files to attach.
    """
    def __init__(self, sender: address.Address, to: List[address.AddressField],
                 smtp_server: str, tenant_id: uuid.UUID, port: int,
                 attach: List[str]) -> None:
        self.sender = sender
        self.to = to
        self.smtp_server = smtp_server
        self.tenant_id = tenant_id
        self.port = port
        self.attach = attach

    @classmethod
    def load(cls, filepath: str) -> Manifest:
        """Create a Manifest from a YAML file on disk.

        Args:
            filepath (str): The path to the YAML file to use.
        """
        # try and load, print any errors if they occur
        with open(filepath, "r") as manifest_file:
            logging.info(f"Loading manifest file: {filepath}")
            manifest_yaml = yaml.safe_load(manifest_file)
            try:
                return MANIFEST_SCHEMA.load(manifest_yaml)
            except ValidationError as err:
                __print_validation_err(err, filepath)


def __print_validation_err(err: ValidationError, manifest_name: str) -> None:
    """Internal function used for logging a validation error in the Manifest
    Schema.

    Args:
        err (ValidationError): The error to log.
        manifest_name (str): A human-readable identifier for the manifest.
    """
    # build up a string for each error
    log_str = []
    log_str.append(f"Error loading manifest '{manifest_name}':")
    for field_name, err_msgs in err.messages.items():
        log_str.append(f"{field_name}: {err_msgs}")

    # log the joined up string, and exit with an error
    logging.critical(" ".join(log_str))
    raise SystemExit(1)


def send_manifest(manifest: Manifest) -> None:
    """Send a given manifest's email to its target.

    Args:
        manifest: The manifest to send.
    """
    msg = message.Message(to=manifest.to,
                          from_addresses=[manifest.sender],
                          date=datetime.now())
    for attachment in manifest.attach:
        if not path.exists(attachment):
            logging.error(f"Unable to attach {attachment}, file did not exist")
            return
        logging.info(f"Attaching {attachment}...")
        msg.attach(attachment)
    msg.headers["X-FileTrust-Tenant"] = str(manifest.tenant_id)

    smtp_url = f"{manifest.smtp_server}:{manifest.port}"
    logging.info(f"Sending manifest to: {smtp_url}...")
    smtp.send(msg, smtp_url)


"""config.py

Config is used for loading the YAML config file used with the script.

Author:
    Sam Gibson <sgibson@glasswallsolutions.com>

Date:
    2019-09-12
"""
import logging
from os.path import basename
from typing import Dict, List

from marshmallow import Schema, fields, post_load, EXCLUDE, ValidationError
from victoria.encryption.schemas import EncryptionEnvelopeSchema, EncryptionEnvelope
import yaml


class MailToilConfigSchema(Schema):
    """ConfigSchema is a marshmallow schema used for validating loaded configs.

    Fields:
        service_bus_connection_strings (fields.Dict[str, EncryptionEnvelope]): Mapping of cluster names to service bus connection strings.
        queues (fields.List[str]): List of queue names to search for dead letters.
        storage_connection_strings (fields.Dict[str, EncryptionEnvelope]): Mapping of cluster names to storage accounts.
        vault_dir (fields.Str): The path to the vault.
    """
    service_bus_connection_strings = fields.Dict(keys=fields.Str(),
                                                 values=fields.Str(),
                                                 required=True)
    queues = fields.List(fields.Str(), required=True)
    storage_accounts = fields.Dict(keys=fields.Str(),
                                   values=fields.Str(),
                                   required=True)
    vault_dir = fields.Str(required=True)

    @post_load
    def make_config(self, data, **kwargs):
        """Callback used by marshmallow after loading object. We're using it here
        to create an instance of Config after loading the data."""
        return MailToilConfig(**data)


CONFIG_SCHEMA = MailToilConfigSchema(unknown=EXCLUDE)
"""Instance of ConfigSchema to use for validation."""


class MailToilConfig:
    """Config is used to store mailreconstruction config.

    Attributes:
        service_bus_connection_strings (Dict[str, str]): Mapping of cluster names to service bus connection strings.
        queues (List[str]): List of queue names to search for dead letters.
        storage_connection_strings (Dict[str, str]): Mapping of cluster names to storage account connection strings.
        vault_dir (str): The path to the vault
    """
    def __init__(self,
                 service_bus_connection_strings: Dict[str, EncryptionEnvelope],
                 queues: List[str], storage_accounts: Dict[str, EncryptionEnvelope],
                 vault_dir: str) -> None:
        self.service_bus_connection_strings = service_bus_connection_strings
        self.queues = queues
        self.storage_accounts = storage_accounts
        self.vault_dir = vault_dir

    def get_service_bus_connection_str(self,
                                       cluster: str) -> EncryptionEnvelope:
        """Try and get a service bus connection string for a cluster.

        Args:
            cluster (str): The cluster name. Should be as defined in config file.

        Returns:
            EncryptionEnvelope: The encrypted service bus connection string.

        Raises:
            ValueError: if the cluster name wasn't found in the config.
        """
        if not self.service_bus_connection_strings.get(cluster):
            raise ValueError(
                f"Couldn't find service bus connection string for cluster '{cluster}'"
            )
        return self.service_bus_connection_strings[cluster]

    def get_storage_account(self, cluster: str) -> EncryptionEnvelope:
        """Try and get a storage connection string for a cluster.

        Args:
            cluster (str): The cluster name. Should be as defined in config file.

        Returns:
            EncryptionEnvelope: The encrypted storage connection string.

        Raises:
            ValueError: if the cluster name wasn't found in the config.
        """
        if not self.storage_accounts.get(cluster):
            raise ValueError(
                f"Couldn't find storage account for cluster '{cluster}'")
        return self.storage_accounts[cluster]



"""load_test

The functionality of the load testing command.
Allows sending emails to an endpoint at a specified rate for a specified time.

Interfaces with an Azure function on the backend (https://github.com/glasswall-sre/going-postal)
that exposes an HTTP endpoint for sending mail.

The backend will return HTTP 200 in the event of a successful send, and 5xx
otherwise.

Author:
    Sam Gibson <sgibson@glasswallsolutions.com>
"""
import asyncio
import random

from dataclasses import dataclass
from datetime import datetime
from functools import reduce
from itertools import cycle

import aiohttp
import roundrobin

from .core.util import generate_random_numbers, generate_random_uuids
from .schemas import LoadTestConfig, Distribution, Function


@dataclass
class TestResult:
    """Data class to store the result of a single test.

    Attributes:
        status (int): The HTTP status code returned from the Azure function.
        message (str): The message returned from the Azure function.
        time (datetime): The time this test result was created.
        function_endpoint (Function): function endpoint used
    """
    status: int
    message: str
    time: datetime
    function_endpoint: Function


async def run_single_test(session: aiohttp.ClientSession, endpoint: str,
                          recipient: str, sender: str, function_endpoint: Function,
                          load_test_config: LoadTestConfig) -> TestResult:
    """Asynchronously run a single test by async invoking the backend Azure
    function. Return a TestResult object containing the results of the test.

    Args:
        session: The opened aiohttp session used for all requests.
        endpoint: The SMTP endpoint to send mail to.
        recipient: The email address to send to.
        sender: The email address to send from.
        function_endpoint: Function endpoint and function code
        load_test_config: The config of the load tester.

    Returns:
        TestResult: The result of the test.
    """
    # extract the endpoint and port from the given endpoint
    split_endpoint = endpoint.split(":")
    endpoint = split_endpoint[0]
    port = split_endpoint[1] if len(
        split_endpoint) > 1 else "25"  # default port 25

    # build the request body with the data provided

    req_body = {
        "endpoint": endpoint,
        "port": int(port),
        "tenant_ids": [str(tenant_id) for tenant_id in load_test_config.tenant_ids],
        "recipient": recipient,
        "sender": sender,
        "timeout": load_test_config.timeout,
        "load": {
            "distribution": [
                {
                    "file": str(distribution.file),
                    "weight": int(distribution.weight)
                }
                for distribution in load_test_config.load.distribution
            ],
            "attachment_count": [int(x) for x in load_test_config.load.attachment_count]
        }
    }

    # perform the POST to run the test
    try:
        async with session.post(
                function_endpoint.function,
                json=req_body,
                params={"code": function_endpoint.mail_send_function_code},
                timeout=aiohttp.ClientTimeout(total=None)) as resp:
            time_now = datetime.now()
            return TestResult(resp.status, await resp.text(), time_now, function_endpoint)
    except aiohttp.ClientConnectorError as error:
        time_now = datetime.now()
        return TestResult(504, str(error), time_now, function_endpoint)

    except Exception as error:
        return TestResult(0, str(error), time_now, function_endpoint)


async def perform_load_test(frequency: int, endpoint: str, duration: int,
                            recipient: str, sender: str, tenant_ids: list,
                            load_test_config: LoadTestConfig) -> None:
    """Asynchronously run a load test with a given frequency and duration.

    Args:
        frequency: How many tests to run per second.
        endpoint: The SMTP endpoint to send to.
        duration: How long in seconds to run the test.
        recipient: The email address to send to.
        sender: The email address to send with.
        tenant_ids: Tenant ids.
        load_test_config: The config for the load tester.
    """
    async with aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(
            total=None)) as session:
        # between each test we wait for 1/frequency seconds
        wait_interval = 1.0 / frequency
        if not tenant_ids:
            tenant_ids = load_test_config.tenant_ids if len(
                load_test_config.tenant_ids) > 0 else generate_random_uuids()
        distributions = Distribution.get_random_distributions()
        load_test_config.load.attachment_count = load_test_config.load.attachment_count if None else generate_random_numbers()
        load_test_config.tenant_ids = tenant_ids

        # perform the tests spread out along the time period

        successful_tests = 0
        unresolvable_exceptions = []
        total_tests = int(frequency * duration)
        number_of_intervals = total_tests
        errors = True
        retry_count = 0
        max_retries = len(load_test_config.mail_send_function_endpoints)
        while errors and retry_count < max_retries:
            tasks = []
            if load_test_config.mail_send_function_endpoints:
                # functions_cycle = cycle(load_test_config.mail_send_function_endpoints)
                intervals_processed = 0
                functions_cycle = roundrobin.basic(load_test_config.mail_send_function_endpoints)
                while intervals_processed < number_of_intervals:
                    # Create a task to run a single test asynchronously
                    function_endpoint = functions_cycle()
                    load_test_config.load.distribution = [random.choice(distributions)]
                    tasks.append(
                        asyncio.create_task(
                            run_single_test(
                                session,
                                endpoint,
                                recipient,
                                sender,
                                function_endpoint,
                                load_test_config
                            )
                        )
                    )

                    intervals_processed += 1
                    percent_done = (intervals_processed / number_of_intervals) * 100
                    print(f"{intervals_processed} / {number_of_intervals}\t\t\t\t{percent_done:02f}%")
                    await asyncio.sleep(wait_interval)

            else:
                print("\n\nFunction endpoints aren't available")

            print("\n")
            # Wait for the tasks to complete
            if not tasks:
                errors = False
                continue
            done, _ = await asyncio.wait(tasks, return_when=asyncio.FIRST_EXCEPTION)
            test_results = []

            for task in done:
                # If there were any exceptions, raise the first one
                error = task.exception()
                if error is not None:
                    raise error
                # Otherwise add the result to the list
                test_results.append(task.result())

            # Compile the amount of successful tests to show the user
            num_successful = reduce(lambda cur, result: cur + (1 if result.status == 200 else 0), test_results, 0)
            successful_tests += num_successful
            print(f"{num_successful} / {number_of_intervals} tests were successfully sent")

            # If there were any non-successful tests, print the reasons
            errors_range = list(range(400, 500)) + [504]
            if num_successful < number_of_intervals:

                failed_sends = [result for result in test_results if result.status != 200]
                number_of_intervals = len(
                    [test for test in failed_sends if test.status in errors_range])

                unresolvable_exceptions += [test for test in failed_sends if test.status not in errors_range]
                removed_endpoints = []
                # FOR DEBUG
                # print("\nReasons for failures:")
                for failed_result in failed_sends:
                    # FOR DEBUG
                    # print(
                    #     f"\t{failed_result.time.isoformat()} - {failed_result.status} error - {failed_result.message}")
                    # print(
                    #     f"\t{failed_result.time.isoformat()} - {failed_result.status} status ")
                    # if not failed_result.message:
                    #     print(failed_result)

                    if failed_result.status in errors_range and failed_result.function_endpoint in load_test_config.mail_send_function_endpoints:
                        load_test_config.mail_send_function_endpoints.remove(failed_result.function_endpoint)
                        removed_endpoints.append(failed_result.function_endpoint)
                print("\n")
                for function_endpoint in removed_endpoints:
                    print(f'{function_endpoint.function} - Failed to send tests')
                if not number_of_intervals:
                    break
            else:
                errors = False

        print(f"total {successful_tests} / {total_tests} tests were successfully sent")

        # FOR DEBUG
        if len(unresolvable_exceptions):
            print("\nUnresolvable failures:")

            for failed_result in unresolvable_exceptions:
                print(f"\t{failed_result.time.isoformat()} - {failed_result.status} error - {failed_result.message}")

        if not retry_count < max_retries:
            print('\nMax retries exceeded')



"""replay_deadletters

The functionality for the replay command from MailToil is implemented here.

Author:
    Sam Gibson <sgibson@glasswallsolutions.com>
"""
import logging
import logging.config
import os
from os.path import join
from typing import List

from .core import config
from .core import service_bus
from .schemas import EmailConfig


def create_vault(queues: List[str], cluster: str, vault_dir: str) -> None:
    """Create the vault. The vault is used to store message bodies in the event
    of a failure when replaying them. This ensures you don't lose messages.

    The vault structure is:
    vault-directory
        cluster0
            queue0
            queue1
            ...
            queueN
        cluster1
        ...
        clusterN

    Args:
        queues (List[str]): The list of queues to create vaults for.
        cluster (str): The cluster to create a vault for.
        vault_dir (str): The root directory of the vault.
    """
    logging.info(f"Creating vault for {cluster}...")

    # for each queue, create a directory
    for queue in queues:
        try:
            os.makedirs(join(vault_dir, cluster, queue))
        except FileExistsError:
            # if it already exists do nothing
            pass


def replay(cfg: config.MailToilConfig, cluster: str, plugin_cfg: EmailConfig):
    """Perform the mailtoil replay functionality.

    Args:
        cfg: The mail toil config.
        cluster: The cluster to replay from.
        plugin_cfg: The email plugin config.
    """
    #encryption_provider = plugin_cfg.victoria_config.get_encryption()
    #service_bus_conn_str = encryption_provider.decrypt_str(
    #    cfg.get_service_bus_connection_str(cluster))
    service_bus_conn_str = cfg.get_service_bus_connection_str(cluster)

    if service_bus_conn_str is None:
        raise SystemExit(1)

    # create the vault
    create_vault(cfg.queues, cluster, cfg.vault_dir)

    # connect to the service bus
    client = service_bus.connect(service_bus_conn_str)

    # scan for and resend dead letters on each queue in the config
    resent_transaction_ids = []
    logging.info(f"Scanning queues on '{cluster}' for dead letters...")
    for queue in cfg.queues:
        vault_dir = join(cfg.vault_dir, cluster, queue)
        resent = service_bus.resend_dead_letters_from_queue(
            vault_dir, queue, client)
        resent_transaction_ids += resent

    # print resent transaction IDs so an SRE can go and check if they were sent
    logging.info(f"{len(resent_transaction_ids)} dead letters were resent." \
        " Please check if the following transaction IDs were successfully relayed.")
    for tx_id in resent_transaction_ids:
        logging.info(tx_id)


from setuptools import setup, find_packages


def repo_file_as_string(file_path: str) -> str:
    with open(file_path, "r") as repo_file:
        return repo_file.read()


setup(install_requires=[
    "victoria", "aiohttp>=3.5.4", "aiorun", "cchardet", "aiodns",
    "azure-servicebus<=0.50.1", "azure-storage-blob>=2.1.0", "sremail", "roundrobin"
],
      name="victoria_email",
      version="v0.6.2",
      description="Victoria plugin for Glasswall Rebuild for Email",
      long_description=repo_file_as_string("README.md"),
      long_description_content_type="text/markdown",
      author="Sam Gibson",
      author_email="sgibson@glasswallsolutions.com",
      url="https://github.com/glasswall-sre/victoria_email",
      packages=find_packages("."),
      python_requires=">=3.7")



import base64
import email
from email import generator
from email.message import MIMEPart
from email.mime.message import MIMEMessage
import logging
from os.path import join
import re
from typing import List, Union, Tuple, Iterator

ANONYMISED_SUFFIX = "_anon"
"""The suffix to put on a file that has been anonymised."""

EMAIL_PATTERN = "<(.*?)>"
"""A regex pattern to get the RFC5322 addr-spec from an email address string.
Even with an optional display-name present."""


def preprocess_mime_message(mime_message: str) -> str:
    """The MIME message contained in the blob storage has characters escaped, as
    well as quotation marks surrounding. This is gonna mess up the MIME parsing,
    so we want to make sure this stuff is all straightened out.

    Args:
        mime_message (str): The MIME message to process.

    Returns:
        str: The preprocessed MIME message.
    """
    # trim quotation marks surrounding
    mime_message = mime_message[1:len(mime_message) - 1]

    # unescape escaped chars
    mime_message = mime_message.replace("\\n", "\n").replace(
        "\\\"", "\"").replace("\\t", "\t").replace("3D\"",
                                                   "\"")  #.replace("=\n", "")

    return mime_message


def write_message(output_dir: str,
                  transaction_id: str,
                  message: MIMEMessage,
                  suffix: str = "") -> None:
    """Write a message to an EML file in a given location.

    This is useful for if you want to open a reconstructed email in outlook.

    Args:
        output_dir (str): The directory to write the file to.
        transaction_id (str): The transaction ID of the email.
        message (MIMEMessage): The parsed MIME message to write.
        suffix (str): The suffix to add to the filename.
    """
    filename = transaction_id + suffix + ".eml"
    logging.info(f"    Writing message '{filename}'")
    with open(join(output_dir, filename), "w") as eml_file:
        gen = generator.Generator(eml_file)
        gen.flatten(message)


def write_attachment(output_dir: str, transaction_id: str,
                     part: MIMEPart) -> None:
    """Write an attachment file to disk.

    Args:
        output_dir (str): The directory to write the file to.
        transaction_id (str): The transaction ID of the email.
        part (MIMEPart): The MIME part representing the attachment.
    """
    filename = f"{transaction_id}_{part.get_filename()}"
    logging.info(f"    Writing attachment '{filename}'")
    with open(join(output_dir, filename), "wb") as attachment_file:
        decoded = base64.b64decode(part.get_payload())
        attachment_file.write(decoded)


def write_html(output_dir: str, transaction_id: str, part: MIMEPart) -> None:
    """Write the HTML part of an email to disk.

    Args:
        output_dir (str): The directory to write the file to.
        transaction_id (str): The transaction ID of the email.
        part (MIMEPart): The MIME part representing the attachment.
    """
    filename = f"{transaction_id}_html.html"
    logging.info(f"    Writing email body '{filename}'")
    with open(join(output_dir, filename), "w") as body_file:
        # the replace is due to weird formatting of HTML in MIME
        body_file.write(part.get_payload().replace("=\n", ""))


def anonymise_mime_message(msg: MIMEMessage) -> MIMEMessage:
    """Replace all of the sensitive fields in the MIME message with messages
    indicating that this message has been anonymised.

    The fields that get replaced are "Subject", "Thread-Topic", as well as the
    body of the email. Attachments are preserved.
    """
    msg.replace_header("Subject", "The subject has been removed for anonymity")

    # the thread topic isn't always present
    if msg.get("Thread-Topic") is not None:
        msg.replace_header("Thread-Topic",
                           "The topic has been removed for anonymity")

    for part in msg.walk():
        # the only parts that contain the body are text/html and text/plain
        # TODO(sam): I think sometimes these can be encoded as base64?
        #            Might need special handling...
        if part.get_content_type() == "text/html" \
                or part.get_content_type() == "text/plain":
            if not is_file(part):
                part.set_payload("The body has been removed for anonymity")

    return msg


def is_file(part: MIMEPart) -> bool:
    """Return True if this MIMEPart is attached as a file."""
    disposition = part.get_content_disposition()
    return disposition == "inline" or disposition == "attachment"


def is_html(part: MIMEPart) -> bool:
    """Return True if this MIMEPart is HTML."""
    return part.get_content_type() == "text/html"


def write_mime_message_payloads(
    output_dir: str,
    transaction_id: str,
    msg: MIMEMessage,
):
    """For each payload in the MIME message, write out the ones that are
    attached as files.

    Args:
        msg (MIMEMessage): The parsed MIME message.
        transaction_id (str): The transaction ID of the email.
        output_dir (str): The directory to write the payloads in.
    """
    for part in msg.walk():
        if is_html(part) and not is_file(part):
            write_html(output_dir, transaction_id, part)

            # for some reason, outlook doesn't like EML files with HTML in them,
            # so we blank out the HTML payload here to display it correctly
            part.set_payload("")
        elif is_file(part):
            write_attachment(output_dir, transaction_id, part)


def process_mime_message(mime_message: str,
                         transaction_id: str,
                         output_dir: str,
                         anonymise: bool = False) -> None:
    """Process a MIME message, writing it and its attachments to files.

    Args:
        mime_message (str): The MIME message as a string.
        transaction_id (str): The transaction ID of the email.
        output_dir (str): The dir to put files in.
        anonymise (bool): Whether to write an anonymised version of the email too.
    """
    logging.info(f"--> Processing message for transaction '{transaction_id}'")
    preprocessed_msg = preprocess_mime_message(mime_message)
    parsed_msg = email.message_from_string(preprocessed_msg)

    # write the email to a file
    write_message(output_dir, transaction_id, parsed_msg)

    # write the email attachments to files
    write_mime_message_payloads(output_dir, transaction_id, parsed_msg)

    if anonymise:
        # anonymise the message and write it to the output dir
        anonymised_msg = anonymise_mime_message(parsed_msg)
        write_message(output_dir,
                      transaction_id,
                      anonymised_msg,
                      suffix=ANONYMISED_SUFFIX)


def extract_emails(addrs: str) -> List[str]:
    return re.findall(EMAIL_PATTERN, addrs)


def filter_by_organisation(addrs: Iterator[str],
                           organisation: str) -> Iterator[str]:
    return [addr for addr in addrs if addr.endswith(organisation)]


def normalise_addresses(addrs: Iterator[str]) -> Iterator[str]:
    return [addr.lower() for addr in addrs]


def get_to_addresses(
        mime_message: str,
        organisation_url: Union[str, None] = "") -> Tuple[List[str], bool]:
    if organisation_url is None:
        organisation_url = ""

    preprocessed_msg = preprocess_mime_message(mime_message)
    parsed_msg = email.message_from_string(preprocessed_msg)

    # extract the to addresses
    to_emails = []
    if parsed_msg.get("To") is not None:
        to_emails += extract_emails(parsed_msg["To"])

    # extract the Cc addresses
    if parsed_msg.get("Cc") is not None:
        to_emails += extract_emails(parsed_msg["Cc"])

    # extract the from addresses
    from_emails = []
    if parsed_msg.get("From") is not None:
        from_emails += extract_emails(parsed_msg["From"])

    # filter them by organisation, if the length after filter is > 0, then it's internal
    # as we already know at least one recipient is in the organisation (or we wouldn't have it!)
    from_emails = filter_by_organisation(from_emails, organisation_url)
    is_internal = len(from_emails) > 0

    # filter by the organisation URL
    return (normalise_addresses(
        filter_by_organisation(to_emails, organisation_url)), is_internal)
